
(*

	Form_GIF unit v1.00 ( for TMT graph )

	// Algorithms in C	by	Sverre Huseby
	// Upgraded&modified	by	Ruzaev Sergei
	// e-mail : druiman@chat.ru

	Can loads :

	01) 4 bit gif	-> 8,16 bit page
	02) 8 bit gif	-> 8,16 bit page
	
	Can saves :

	01) 8 bit page	-> 8 bit gif

*)


unit Form_GIF;

interface

const
	-- Positions
	Normal			=  0;
	Center			=  1;

	-- Error Codes
	ImageOk			=  0;
	BadFile			=  1;
	BadGifCode		=  2;
	NotAGifFile		=  3;
	InvalidBlockSize	=  4;
	BadSymbolSize		=  5;
	BadFirstGifCode		=  6;
	BadSeek			=  7;
	UnknownError		=255;

	-- Start Pos
	fromX			=  0;
	fromY			=  0;
	userSeek		=  0;

var
GifPal				: array[0..767] of byte;
GifError			: integer;
lastGifDX			: longint;
lastGifDY			: longint;
lastGifX0			: longint;
lastGifY0			: longint;

function  lastError: string;
procedure loadGif(filename: string;position: byte;fseek: dword);
function  saveGif(filename : string;x0,y0,x1,y1: longint): longint;

implementation

uses Graph;

type
AnyImage=record
	Width,Height		: word;
	BitsPerPixel		: byte;
	NumberOfColors		: word;
	SizeOfImage		: longint;
	Information		: word;
end;
GifImageHeader=record
	Signature		: array [1..6] of char;
	ScreenWidth,ScreenHeight: word;
	flags,background,aspect : byte;
end;

function safeshr(value: longint;cnt: word): longint; assembler;
asm
	mov	eax,value
	mov	bx,ax
	shr	eax,16
	mov	dx,ax
	mov	ax,bx
	mov	cx,cnt
	shrd	ax,dx,cl
	shr	dx,cl

	mov	bx,ax
	mov	ax,dx
	shl	eax,16
	mov	ax,bx
end;

function safeshl(value: longint;cnt: word): longint; assembler;
asm
	mov	eax,value
	mov	bx,ax
	shr	eax,16
	mov	dx,ax
	mov	ax,bx
	mov	cx,cnt
	shld	dx,ax,cl
	shl	ax,cl

	mov	bx,ax
	mov	ax,dx
	shl	eax,16
	mov	ax,bx
end;

function  lastError: string;
begin
	case GifError of
	 ImageOk		: lastError:='all ok';
 	 BadFile		: lastError:='bad GIF file';
	 BadGifCode		: lastError:='bad GIF code';
	 NotAGifFile		: lastError:='not a GIF file';
	 InvalidBlockSize	: lastError:='invalid GIF block size';
	 BadSymbolSize		: lastError:='bad GIF symbol table';
	 BadFirstGifCode	: lastError:='bad first GIF code';
	 else			  lastError:='unknown error';
	end;
end;

procedure loadGif(filename: string;position: byte;fseek: dword);

type
SomeImageBlock=record
	left,top,width,height	: word;
	flags			: byte;
end;

var
	f			: file;
	GifImage		: AnyImage;
	WorkSpace1, WorkSpace2  : array[0..4095] of byte;
	WorkSpace3              : array[0..4095] of integer;
	StringHashtable         : array[0..1] of integer;
	Interlaced              : boolean;
	ImageWidth, ImageHeight : integer;
	ImageColorDepth         : byte;
	ImagePaletteSize        : integer;
	ColorBits               : byte;
	GifBlock                : array [0..255] of byte;
	GifBlockindex           : word;
	NumStrings              : word;
	GifHeader               : GifImageHeader;
	ImageBlock              : SomeImageBlock;
	ImageSize		: longint;
	ja			: word;
	line_nb			: integer;
	line_gif		: array[0..1023] of byte;
	currentX		: integer;
	maxColor		: longint;
	tabTable		: array[0..255] of word;

-- r,g,b color tab to 65536 word color
function tab16(r,g,b: byte): word; assembler;
asm
	MOV	DH,R
	SHR	DH,1
	MOV	DL,G
	SHL	DL,2
	SHL	DX,3
	MOV	CL,B
	SHR	CL,1
	ADD	DL,CL
	MOV	AX,DX
end;

function get(var x;nb : word) : boolean;
var 
j	: word;
begin
 blockread(f, x, nb, j);
 if ioresult<>0
  then begin
  get:=false;
  exit;
  end;
 if j<>nb then get:=false else get:=true;
end;

procedure write_linepos(x0,y0,xd: word);
var
xCross		: word;
begin
for xCross:=x0 to x0+xd-1 do			-- output bf to screen
 begin
 putpixel(lastGifX0+xCross,lastGifY0+y0,tabTable[line_gif[xCross-x0]]);
 end;
end;

function put(x,j : integer) : boolean;
var
n		: word;
data		: array[0..1] of byte absolute x;
begin
n:=0;
while (n<j) do
 begin
 line_gif[currentx]:=data[n];
 inc(n);
 inc(currentX);
 if currentX>=gifimage.width then 
  begin
  write_linepos(fromx, fromy+line_nb, gifimage.width);
  currentX:=0;
  end;
 end;
 put:=true;
end;

procedure UnpackImage(bits : integer);

type
 TByteArray = Array [0..0] Of byte;

Var
  bits2, codesize, codesize2, nextcode, thiscode,
  oldtoken, currentcode, oldcode, bitsleft, blocksize,
  pass, byt, p, q, u, i : integer;
  ScanlineDataPos : LongInt;

Const
  wordmasktable : Array [0..15] Of Word =
	  ( $0000, $0001, $0003, $0007,
            $000F, $001F, $003F, $007F,
            $00FF, $01FF, $03FF, $07FF,
            $0FFF, $1FFF, $3FFF, $7FFF );
  inctable : Array [0..4] Of Integer = ( 8, 8, 4, 2, 0 );
  starttable : Array [0..4] Of Integer = ( 0, 4, 2, 1, 0 );


Function GetBlock: Boolean;
Begin
  GetBlock := False;
  blocksize:=0;
  If Not Get(blocksize,1) then
     Begin
       GifError := BadFile;
       Exit;
     end;
  If blocksize>0 then
     Begin
       p:=0;
       If Not Get (GifBlock,blocksize) then
          Begin
            GifError := BadFile;
            Exit;
          end;
       q:=blocksize;
       GetBlock := True;
     End
  Else
     Begin
       GifError:=InvalidBlockSize;
       Exit;
     End;
end;

Begin
  pass:=0;
  line_nb:=0;
  byt:=0;
  p:=0;
  q:=0;
  blocksize:=0;
  FillChar (GifBlock, 256, 0);
  bitsleft:=8;
  If (bits<2) Or (bits>8) then Begin
    GifError := BadSymbolSize;
    Exit;
  End;
  bits2    :=1 SHL bits;
  nextcode :=bits2 + 2;
  codesize :=bits + 1;
  codesize2:=1 SHL codesize;
  oldcode  :=-1;
  oldtoken :=-1;
  while True do begin
    if bitsleft=8 then begin
      Inc (p);
      If p>=q then GetBlock;
      If GifError<>ImageOk then
         Begin
           Exit;
         end;
      bitsleft:=0;
    end;
    thiscode:=GifBlock[p];
    currentcode:=codesize+bitsleft;
    If currentcode<=8 then begin
      GifBlock[p]:=safeshr(GifBlock[p],codesize);
      bitsleft:=currentcode;
    end else begin
      Inc (p);
      If p>=q then
         If Not GetBlock then
            Begin
              Exit;
            end;
      thiscode:=thiscode Or safeshl(GifBlock[p],(8-bitsleft));
      If currentcode<=16 then
        bitsleft:=currentcode-8
      else begin
        Inc (p);
        If p>=q then
           If Not GetBlock then Exit;
        thiscode:=thiscode Or safeshl(GifBlock[p],(16-bitsleft));
        bitsleft:=currentcode-16;
      end;
      GifBlock[p]:=safeshr(GifBlock[p],bitsleft);
    end;
    thiscode:=thiscode and wordmasktable[codesize];
    currentcode:=thiscode;
    If thiscode=bits2+1 then Break;
    If thiscode > nextcode then Begin
      GifError:= BadGifCode;
      Exit;
    End;
    If thiscode=bits2 then Begin
      nextcode:=bits2+2;
      codesize:=bits+1;
      codesize2:=1 SHL codesize;
      oldtoken:=-1;
      OldCode:=-1;
      Continue;
    End;
    u:=0;
    If thiscode=nextcode then Begin
      If oldcode=-1 then Begin
        GifError:=BadFirstGifCode;
        Exit;
      End;
      WorkSpace1[u]:=oldtoken;
      Inc (u);
      thiscode:=oldcode;
    End;
    While thiscode>=bits2 Do Begin
      WorkSpace1[u]:=WorkSpace2[thiscode];
      Inc (u);
      thiscode:=WorkSpace3[thiscode];
    End;
    oldtoken:=thiscode;

    while True do begin
      If Not Put (thiscode,1) then
         Begin
           GifError := BadGifCode;
           Exit;
         end;
      Inc (byt);
      if byt>=ImageWidth then begin
        byt:=0;
        if Interlaced then begin
          line_nb:=line_nb+inctable[pass];
          if line_nb >= ImageHeight then begin
            Inc (pass);
            line_nb:=starttable[pass];
          end;
        end
        else
          Inc (line_nb);
      end;
      if u<=0 then Break;
      Dec (u);
      thiscode:=WorkSpace1[u];
    end;
    If (nextcode<4096) And (oldcode<>-1) then Begin
      WorkSpace3[nextcode]:=oldcode;
      WorkSpace2[nextcode]:=oldtoken;
      Inc (nextcode);
      If (nextcode>=codesize2) And (codesize<12) then Begin
        Inc (codesize);
        codesize2:=1 SHL codesize;
      End;
    End;
    Oldcode:=currentcode;
  End;
End;

procedure SkipExtension;

var
 	n, c	: byte;

begin
	get(c,1);
	case c of
	 $01 : begin
	       get(GifBlock,13);
               get(n,1);
               while n>0 do
		begin
		get(GifBlock,n);
		get(n,1);
		end;
	       end;
	 $F9 : get(GifBlock,6);
	 $FE : begin
	       get(n,1);
	       while n>0 do 
	   	begin
		get(GifBlock,n);
		get(n,1);
		end;
	       end;
	 $FF : begin
	       get(GifBlock,12);
	       get(n,1);
	       while n>0 do 
		begin
		get(GifBlock,n);
		get(n,1);
		end;
	 end;
	 else begin
	      get(n,1);
	      get(GifBlock,n);
	      end;
	end;
end;

procedure GenTabTable;				-- color val set
var
i		: word;
begin
	if MaxColor>255 then			-- to 65536 scr
	begin
	for i:=0 to 255 do
	 tabTable[i]:=tab16(GifPal[i*3+0],
	  	            GifPal[i*3+1],
	                    GifPal[i*3+2]);
	end else
	for i:=0 to 255 do
	 tabTable[i]:=i;
end;

var
	backGround	: byte;
	c		: integer;
	ch		: char;
	i		: word;
	size		: longint;

begin
	CurrentX:=0;
	GifError:=ImageOk;
	MaxColor:=GetMaxColor;			-- What is current scr ?

	assign(f,filename);			-- Open GIF file
  {$i-} reset(f,1); {$i+}
	if ioResult<>0 then			-- IO check
	 begin
	 GifError:=BadFile;
	 exit;
	 end;
  {$i-} seek(f,fseek); {$i+}
	if ioResult<>0 then			-- IO check
	 begin
	 GifError:=BadSeek;
	 exit;
	 end;

	seek(f,userSeek);			-- GIF seek

						-- load Header
	if not get(gifHeader,sizeof(gifHeader)) then
	 begin
	 GifError:=BadGifCode;
	 close(f);
	 exit;
	 end;

	with GifHeader do			-- check for real GIF
	 begin
	 if Signature[1]+Signature[2]+Signature[3]<>'GIF' then
	  begin
	  GifError:=NotAGifFile;
	  close(f);
	  exit;
	  end;
         end;

	ColorBits:=GifHeader.Flags and 7+1;	-- gen bits value
	BackGround:=GifHeader.BackGround;

	
	if GifHeader.Flags And 128=128 then	-- load palette
	begin
	case ColorBits of
	4 : get(GifPal,48);
	8 : get(GifPal,768);
	end;
	end;

	for i:=0 to 767 do			-- RGB to HLS
	 GifPal[i]:=GifPal[i] shr 2;

	GenTabTable;

	with GifHeader do			-- Get Image Size
	begin
	lastGifDX := ScreenWidth;
	lastGifDY := ScreenHeight;
	if position=center then
	 begin
	 lastGifX0 := Succ(GetMaxX) div 2-lastGifDX div 2;
	 lastGifY0 := Succ(GetMaxY) div 2-lastGifDY div 2;
	 if lastGifX0+lastGifDX>GetMaxX then dec(lastGifX0);
	 if lastGifY0+lastGifDY>GetMaxY then dec(lastGifY0);
	 end else
	 begin
	 lastGifX0 := 0;
	 lastGifY0 := 0;
	 end;
	size:=ScreenHeight*ScreenWidth;
	ImageSize:=size;
	end;

	ch:=#0;					-- Real algorithm
	while (ch=',') or (ch='!') or (ch=#0) do
	begin
	get(ch,1);
	case ch of
        ',' : begin
              if not get(ImageBlock,SizeOf(ImageBlock)) then
                 begin
		 GifError:=BadGifCode;
		 close(f);
		 exit;
		 end;
              if (ImageBlock.flags and 128)=128 then	-- Load Pal too
                 begin
                 colorBits:=ImageBlock.Flags And 7+1;
		 case colorBits of
		 4 : Get(GifPal,48);
		 8 : Get(GifPal,768);
		 end;
		 if (colorBits=4) or (colorBits=8) then
		   begin
	 	   for i:=0 to 767 do
		    GifPal[i]:=GifPal[i] shr 2;
		   GenTabTable;
		   end;
                 end;
		ImageWidth       := ImageBlock.Width;
              	ImageHeight      := ImageBlock.Height;
              	ImageColorDepth  := ColorBits;
              	ImagePaletteSize := 1 SHL ColorBits;
		with GifImage do
                begin
                  Height        := ImageHeight;
                  Width         :=  ImageWidth;
                  NumberOfColors:= ImagePaletteSize;
                  BitsPerPixel  := ColorBits;
                  SizeOfImage   := ImageSize;
                  If GifHeader.Signature[5] = '7' then	-- Gif 87a
                     Information := 0
                  else
                     Information := 1;			-- Gif 89a
		end;
                Interlaced:=ImageBlock.Flags And 64=64;
		c:=0;
		if not get(c,1) then
                begin
		GifError:=BadGifCode;
		close(f);
		exit;
		end;
		UnpackImage(c);
		end;
	'!' : SkipExtension;
    	end;
	end;
	close(f);				-- All done
end;

function  saveGif(filename : string;x0,y0,x1,y1: longint): longint;
const
     hash_free = $ffff;
     next_first = $ffff;
     bits_per_pixel = $08;

     maxbits = 12;
     maxstr = 1 shl maxbits;

     code_reserved = 2;

     hashsize = 9973;
     hashstep = 2039;

type
     ttabbyte = array[0..0] of byte;
     ptabbyte = ^ttabbyte;

     ttabword = array[0..0] of word;
     ptabword = ^ttabword;

     screendescriptor = record
                              screenwidth, screenheight : word;
                              flag : byte;
                              background, pixelaspectratio : byte;
                        end;
     imagedescriptor = record
                             separator : char;
                             leftposition, topposition : word;
                             width, height : word;
                             flag : byte;
                       end;

var strchr : ptabbyte;
    strnext, strhash : ptabword;
    numstrings : word;

    outfile : file;

    buffer : array[0..255] of byte;
    index, bitsleft : integer;

    bitsperpixel, numcolors : integer;
    screenheight, screenwidth, relpixx, relpixy : integer;

function hash(index, lastbyte : word) : word;
begin
     hash:=( ( safeshl(lastbyte,8) ) XOR index ) mod hashsize;
end;

function create(s : string) : boolean;
begin
     assign(outfile, s);
     rewrite(outfile, 1);
     (*$ifndef iocheckerror*)
     if ioresult<>0
        then begin
             create:=false;
             exit;
        end;
     (*$endif*)
     create:=true;
end;

function write(var buf;len : word) : boolean;
begin
     blockwrite(outfile, buf, len);
     (*$ifndef iocheckerror*)
     if ioresult<>0
        then begin
             write:=false;
             exit;
        end;
     (*$endif*)
     write:=true;
end;

function writebyte(b : byte) : boolean;
begin
     blockwrite(outfile, b, 1);
     (*$ifndef iocheckerror*)
     if ioresult<>0
        then begin
             writebyte:=false;
             exit;
        end;
     (*$endif*)
     writebyte:=true;
end;

function writeword(w : word) : boolean;
begin
     blockwrite(outfile, w, 2);                 (*In original code, it was splitted in 2 byte output to ensure portability*)
     (*$ifndef iocheckerror*)                   (*Was removed since this library is not portable (as is TP) *)
     if ioresult<>0
        then begin
             writeword:=false;
             exit;
        end;
     (*$endif*)
     writeword:=true;
end;

(*Bit file oriented routines*)
procedure initbitfile;
begin
     index:=0;
     buffer[0]:=0;
     bitsleft:=8;
end;

function resetoutbitfile : boolean;
var numbytes : byte;
begin
     if bitsleft=8
        then numbytes:=index
        else numbytes:=index+1;
     if numbytes>0
        then begin
             if not writebyte(numbytes)
                then begin
                     resetoutbitfile:=false;
                     exit;
                end;
             if not write(buffer, numbytes)
                then begin
                     resetoutbitfile:=false;
                     exit;
                end;
             index:=0;
             buffer[0]:=0;
             bitsleft:=8;
     end;
     resetoutbitfile:=true;
end;

function writebits(bits, numbits : integer) : longint;
var bitswritten, numbytes : integer;
begin
     bitswritten:=0;
     numbytes:=255;
     repeat
           if ( (index=254) and (bitsleft=0) ) or (index>254)
              then begin
                   if not writebyte(numbytes)
                      then begin
                           writebits:=-1;
                           exit;
                      end;
                   if not write(buffer, numbytes)
                      then begin
                           writebits:=-1;
                           exit;
                      end;
                   index:=0;
                   buffer[index]:=0;
                   bitsleft:=8;
              end;
           if (numbits<=bitsleft)
              then begin
                   buffer[index]:=buffer[index] or ( safeshl( bits and ( ( safeshl(1,numbits) -1 ) ),( 8 - bitsleft )));
                   inc(bitswritten, numbits);
                   dec(bitsleft, numbits);
                   numbits:=0;
              end else begin
                   buffer[index]:=buffer[index] or ( safeshl( bits and ( ( safeshl(1,bitsleft) -1 ) ),(8 - bitsleft )));
                   inc(bitswritten, bitsleft);
                   bits:=safeshr(bits,bitsleft);
                   dec(numbits, bitsleft);
                   inc(index);
                   buffer[index]:=0;
                   bitsleft:=8;
              end;
     until ( numbits = 0 );
     writebits:=bitswritten;
end;

procedure freestrtab;
begin
     if ( strhash <> nil )
        then begin
             freemem(strhash, hashsize * sizeof(word));
             strhash:=nil;
        end;

     if ( strnext <> nil )
        then begin
             freemem(strnext, maxstr * sizeof(word));
             strnext:=nil;
        end;

     if ( strchr <> nil )
        then begin
             freemem(strchr, maxstr * sizeof(byte));
             strchr:=nil;
        end;
end;

function allocstrtab : boolean;
begin
     getmem(strchr, maxstr * sizeof(byte));
     if strchr=nil
        then begin
             freestrtab;
             allocstrtab:=false;
             exit;
        end;
     getmem(strnext, maxstr * sizeof(word));
     if strnext=nil
        then begin
             freestrtab;
             allocstrtab:=false;
             exit;
        end;
     getmem(strhash, hashsize * sizeof(word));
     if strhash=nil
        then begin
             freestrtab;
             allocstrtab:=false;
             exit;
        end;
     allocstrtab:=true
end;

function addcharstring(index : word;b : byte) : word;
var hshidx : word;
begin
     if numstrings>=maxstr
        then begin
             addcharstring:=$ffff;
             exit;
        end;

     hshidx:=hash(index, b);
     while (strhash^[hshidx]<>$ffff)
           do hshidx:=(hshidx + hashstep) mod hashsize;

     strhash^[hshidx]:=numstrings;
     strchr^[numstrings]:=b;
     if (index <> $ffff)
        then strnext^[numstrings]:=index
        else strnext^[numstrings]:=next_first;

     addcharstring:=numstrings;
     inc(numstrings);
end;

function findcharstring(index : word;b : byte) : word;
var hshidx, nextidx : word;
begin
     if (index=$ffff)
        then begin
             findcharstring:=b;
             exit;
        end;

     hshidx:=hash(index, b);

     nextidx:=strhash^[hshidx];
     while ( nextidx <> $ffff ) do begin
           if (strnext^[nextidx]=index) and (strchr^[nextidx]=b)
              then begin
                   findcharstring:=nextidx;
                   exit;
              end;
           hshidx:=(hshidx+hashstep) mod hashsize;
           nextidx:=strhash^[hshidx];
     end;

     findcharstring:=$ffff;
end;

procedure clearstrtab(codesize : integer);
var q, w : integer;
    wp : ^word;
begin
     numstrings:=0;

     wp:=@(strhash^[0]);
     for q:=0 to hashsize-1 do begin
         wp^:=hash_free;
         inc(wp, sizeof(word));
     end;

     w:=safeshl(1,codesize) + code_reserved;
     for q:=0 to w-1 do addcharstring($ffff, q);
end;

function inputbyte : integer;
var ret : integer;
begin
     if (relpixy>=screenheight)
        then begin
             inputbyte:=-1;
             exit;
        end;
     ret:=getpixel(x0+relpixx, y0+relpixy) and $ff;
     inc(relpixx);
     if relpixx>=screenwidth
        then begin
             relpixx:=0;
             inc(relpixy);
        end;
     inputbyte:=ret;
end;

function writescreendescriptor(var sd : screendescriptor;globalcolortableflag, colorresolution, sortflag,
         globalcolortablesize : integer) : boolean;
begin
     sd.flag:=safeshl(globalcolortableflag,7) or safeshl(colorresolution,4) or safeshl(sortflag,3) or globalcolortablesize;
     blockwrite(outfile, sd, sizeof(screendescriptor));
     (*$ifndef iocheckerror*)
     if ioresult<>0
        then begin
             writescreendescriptor:=false;
             exit;
        end;
     (*$endif*)
     writescreendescriptor:=true;
end;

function writeimagedescriptor(var id : imagedescriptor;localcoltab, interlace, sort, reserved,
         localtabsize : integer) : boolean;
begin
     id.flag:=safeshl(localcoltab,7) or safeshl(interlace,6) or safeshl(sort,5) or safeshl(reserved,3) or localtabsize;
     blockwrite(outfile, id, sizeof(imagedescriptor));
     (*$ifndef iocheckerror*)
     if ioresult<>0
        then begin
             writeimagedescriptor:=false;
             exit;
        end;
     (*$endif*)
     writeimagedescriptor:=true;
end;

function lzw_compress(codesize : integer) : boolean;
var c : integer;
    index : word;
    clearcode, endofinfo, numbits, limit, errcode : integer;
    prefix : word;
begin
     prefix:=$ffff;

     initbitfile;

     clearcode:=safeshl(1,codesize);
     endofinfo:=clearcode + 1;

     numbits:=codesize + 1;
     limit:=safeshl(1,numbits) -1;

     if not allocstrtab
        then begin
             lzw_compress:=false;
             exit;
        end;
     clearstrtab(codesize);

     writebits(clearcode, numbits);

     c:=inputbyte;

     while (c<>-1) do begin
           index:=findcharstring(prefix, c);
           if index<>$ffff
              then prefix:=index
              else begin
                   writebits(prefix, numbits);
                   if ( addcharstring(prefix, c) > limit )
                      then begin
                           inc(numbits);
                           if numbits>12
                              then begin
                                   writebits(clearcode, numbits-1);
                                   clearstrtab(codesize);
                                   numbits:=codesize+1;
                              end;
                           limit:=safeshl(1,numbits)-1;
                      end;

                   prefix:=c;
              end;
           c:=inputbyte;
     end;

     if prefix<>$ffff
        then writebits(prefix, numbits);

     writebits(endofinfo, numbits);

     resetoutbitfile;

     freestrtab;

     lzw_compress:=true;
end;

function bitsneeded(n : word): integer;
var ret : integer;
begin
     ret:=1;
     if n=0
        then begin
             bitsneeded:=0;
             exit;
        end;
     dec(n);
     n:=safeshr(n,1);
     while n>0 do begin
           inc(ret);
           n:=safeshr(n,1);
     end;

     bitsneeded:=ret;
end;

var id : imagedescriptor;
    sd : screendescriptor;
    toto : string[8];
    temp : longint;

begin
     numcolors:=safeshl(1,bitsperpixel);
     screenheight:=y1-y0+1;
     screenwidth:=x1-x0+1;

     if not create(filename)
        then begin
             result:=-1;
             exit;
        end;

     toto:='GIF87a';
     if not write(toto[1], 6)
        then begin
             result:=-1;
             exit;
        end;

     sd.screenwidth:=screenwidth;
     sd.screenheight:=screenheight;
     sd.background:=0;
     sd.pixelaspectratio:=0;
     if not writescreendescriptor(sd, bits_per_pixel-1, 0, bits_per_pixel-1, bits_per_pixel-1)
        then begin
             result:=-1;
             exit;
        end;
     for temp:=0 to 767 do
      gifpal[temp] := gifpal[temp] shl 2;
     if not write(gifpal, 768)
        then begin
             result:=-1;
             exit;
        end;
     id.separator:=',';
     id.leftposition:=0;
     id.topposition:=0;
     id.width:=screenwidth;
     id.height:=screenheight;
     if not writeimagedescriptor(id, 0, 0, 0, 0, 0)
        then begin
             result:=-1;
             exit;
        end;

     if not writebyte(bits_per_pixel)
        then begin
             result:=-1;
             exit;
        end;

     relpixx:=0;
     relpixy:=0;

     if not lzw_compress(bits_per_pixel)
        then begin
             result:=-1;
             exit;
        end;

     if not writebyte(0)
        then begin
             result:=-1;
             exit;
        end;

     id.separator:=';';
     if not writeimagedescriptor(id, 0, 0, 0, 0, 0)
        then begin
             result:=-1;
             exit;
        end;

     result:=filesize(outfile);

     close(outfile);
end;

end.